import struct

from wand.image import Image

from win2xcur.cursor import CursorFrame, CursorImage


class CURParser:
    MAGIC = b'\0\0\02\0'
    ICON_DIR = struct.Struct('<HHH')
    ICON_DIR_ENTRY = struct.Struct('<BBBBHHII')

    @classmethod
    def can_parse(cls, blob):
        return blob[:len(cls.MAGIC)] == cls.MAGIC

    def __init__(self, blob):
        self.blob = blob
        self._image = Image(blob=blob, format='cur')
        self._hotspots = self._parse_header()
        self.frames = [CursorFrame([
            CursorImage(image, hotspot) for image, hotspot in zip(self._image.sequence, self._hotspots)
        ])]

    def _parse_header(self):
        reserved, ico_type, image_count = self.ICON_DIR.unpack(self.blob[:self.ICON_DIR.size])
        assert reserved == 0
        assert ico_type == 2
        assert image_count == len(self._image.sequence)

        offset = self.ICON_DIR.size
        hotspots = []
        for i in range(image_count):
            width, height, palette, reserved, hx, hy, size, file_offset = self.ICON_DIR_ENTRY.unpack(
                self.blob[offset:offset + self.ICON_DIR_ENTRY.size])
            hotspots.append((hx, hy))
            offset += self.ICON_DIR_ENTRY.size

        return hotspots



from win2xcur.parser.ani import ANIParser
from win2xcur.parser.cur import CURParser

PARSERS = [CURParser, ANIParser]


def open_blob(blob):
    for parser in PARSERS:
        if parser.can_parse(blob):
            return parser(blob)
    raise ValueError('Unsupported file format')



from typing import List

from wand.image import BaseImage, Image

from win2xcur.cursor import CursorFrame


def apply_to_image(image: BaseImage, *, color: str, radius: float, sigma: float, xoffset: float,
                   yoffset: float) -> Image:
    opacity = Image(width=image.width, height=image.height, pseudo='xc:white')
    opacity.composite(image.channel_images['opacity'], left=round(xoffset * image.width),
                      top=round(yoffset * image.height))
    opacity.gaussian_blur(radius * image.width, sigma * image.width)
    opacity.negate()
    opacity.modulate(50)

    shadow = Image(width=image.width, height=image.height, pseudo='xc:' + color)
    shadow.composite(opacity, operator='copy_opacity')

    result = image.clone()
    result.composite(shadow, operator='difference')
    return result


def apply_to_frames(frames: List[CursorFrame], **kwargs):
    for frame in frames:
        for cursor in frame:
            cursor.image = apply_to_image(cursor.image, **kwargs)



import struct
from copy import copy

from win2xcur.parser.cur import CURParser


class ANIParser:
    SIGNATURE = b'RIFF'
    ANI_TYPE = b'ACON'
    FRAME_TYPE = b'fram'
    RIFF_HEADER = struct.Struct('<4sI4s')
    CHUNK_HEADER = struct.Struct('<4sI')
    ANIH_HEADER = struct.Struct('<IIIIIIIII')
    UNSIGNED = struct.Struct('<I')
    SEQUENCE_FLAG = 0x2
    ICON_FLAG = 0x1

    @classmethod
    def can_parse(cls, blob):
        signature, size, subtype = cls.RIFF_HEADER.unpack(blob[:cls.RIFF_HEADER.size])
        return signature == cls.SIGNATURE and size == len(blob) - 8 and subtype == cls.ANI_TYPE

    def __init__(self, blob):
        self.blob = blob
        if not self.can_parse(blob):
            raise ValueError('Not a .ani file')
        self.frames = self._parse(self.RIFF_HEADER.size)

    def _unpack(self, struct_cls, offset):
        return struct_cls.unpack(self.blob[offset:offset + struct_cls.size])

    def _read_chunk(self, offset, expected):
        name, size = self._unpack(self.CHUNK_HEADER, offset)
        if name not in expected:
            raise ValueError('Expected chunk %r, found %r' % (expected, name))
        return size, offset + self.CHUNK_HEADER.size

    def _parse(self, offset):
        size, offset = self._read_chunk(offset, expected=[b'anih'])

        if size != self.ANIH_HEADER.size:
            raise ValueError('Unexpected anih header size %d, expected %d' % (size, self.ANIH_HEADER.size))

        size, frame_count, step_count, width, height, bit_count, planes, display_rate, flags = self.ANIH_HEADER.unpack(
            self.blob[offset:offset + self.ANIH_HEADER.size])

        if not flags & self.ICON_FLAG:
            raise NotImplementedError('Raw BMP images not supported.')

        offset += self.ANIH_HEADER.size
        list_size, offset = self._read_chunk(offset, expected=[b'LIST'])
        list_end = list_size + offset

        if self.blob[offset:offset + 4] != self.FRAME_TYPE:
            raise ValueError('Unexpected RIFF list type: %r, expected %r' %
                             (self.blob[offset:offset + 4], self.FRAME_TYPE))
        offset += 4

        frames = []
        for i in range(frame_count):
            size, offset = self._read_chunk(offset, expected=[b'icon'])
            frames.append(CURParser(self.blob[offset:offset + size]).frames[0])
            offset += size

        if offset != list_end:
            raise ValueError('Wrong RIFF list size: %r, expected %r' % (offset, list_end))

        sequence = frames
        if flags & self.SEQUENCE_FLAG:
            size, offset = self._read_chunk(offset, expected=[b'seq '])
            sequence = [copy(frames[i]) for i, in self.UNSIGNED.iter_unpack(self.blob[offset:offset + size])]
            if len(sequence) != step_count:
                raise ValueError('Wrong animation sequence size: %r, expected %r' % (len(sequence), step_count))
            offset += size

        delays = [display_rate for _ in range(step_count)]
        if offset < len(self.blob):
            size, offset = self._read_chunk(offset, expected=[b'rate'])
            delays = [i for i, in self.UNSIGNED.iter_unpack(self.blob[offset:offset + size])]
            if len(sequence) != step_count:
                raise ValueError('Wrong animation rate size: %r, expected %r' % (len(delays), step_count))

        for frame, delay in zip(sequence, delays):
            frame.delay = delay / 60

        return sequence



import os
import subprocess
import sys
from tempfile import TemporaryDirectory
from typing import List

from wand.image import Image

from win2xcur.cursor import CursorFrame

xcursorgen_checked = False


def check_xcursorgen() -> None:
    global xcursorgen_checked
    if xcursorgen_checked:
        return

    try:
        subprocess.check_call(['xcursorgen', '--version'], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    except subprocess.CalledProcessError:
        raise RuntimeError('xcursorgen must be installed to create X11 cursors!')
    else:
        xcursorgen_checked = True


def to_x11(frames: List[CursorFrame]) -> bytes:
    check_xcursorgen()

    counter = 0
    configs = []
    with TemporaryDirectory() as png_dir:
        for frame in frames:
            for cursor in frame:
                name = '%d.png' % (counter,)
                hx, hy = cursor.hotspot
                configs.append('%d %d %d %s %d' % (cursor.image.width, hx, hy, name, int(frame.delay * 1000)))

                image = Image(image=cursor.image)
                image.save(filename=os.path.join(png_dir, name))
                counter += 1

        output_file = os.path.join(png_dir, 'cursor')
        process = subprocess.Popen(['xcursorgen', '-p', png_dir, '-', output_file], stdin=subprocess.PIPE,
                                   stderr=subprocess.PIPE)

        _, error = process.communicate('\n'.join(configs).encode(sys.getfilesystemencoding()))
        if process.wait() != 0:
            raise RuntimeError('xcursorgen failed: %r' % error)

        with open(output_file, 'rb') as f:
            result = f.read()

    return result



import argparse
import os
import sys
import traceback
from multiprocessing import cpu_count
from multiprocessing.pool import ThreadPool
from threading import Lock

from win2xcur import shadow
from win2xcur.parser import open_blob
from win2xcur.writer.x11 import check_xcursorgen, to_x11


def main() -> None:
    parser = argparse.ArgumentParser(description='Converts Windows cursors to X11 cursors.')
    parser.add_argument('files', type=argparse.FileType('rb'), nargs='+',
                        help='Windows cursor files to convert (*.cur, *.ani)')
    parser.add_argument('-o', '--output', '--output-dir', default=os.curdir,
                        help='Directory to store converted cursor files.')
    parser.add_argument('-s', '--shadow', action='store_true',
                        help="Whether to emulate Windows's shadow effect")
    parser.add_argument('-O', '--shadow-opacity', type=int, default=50,
                        help='Opacity of the shadow (0 to 255)')
    parser.add_argument('-r', '--shadow-radius', type=float, default=0.1,
                        help='Radius of shadow blur effect (as fraction of width)')
    parser.add_argument('-S', '--shadow-sigma', type=float, default=0.1,
                        help='Sigma of shadow blur effect (as fraction of width)')
    parser.add_argument('-x', '--shadow-x', type=float, default=0.05,
                        help='x-offset of shadow (as fraction of width)')
    parser.add_argument('-y', '--shadow-y', type=float, default=0.05,
                        help='x-offset of shadow (as fraction of height)')
    parser.add_argument('-c', '--shadow-color', default='#000000',
                        help='color of the shadow')

    args = parser.parse_args()
    print_lock = Lock()

    check_xcursorgen()

    def process(file):
        name = file.name
        blob = file.read()
        try:
            cursor = open_blob(blob)
        except Exception:
            with print_lock:
                print('Error occurred while processing %s:' % (name,), file=sys.stderr)
                traceback.print_exc()
        else:
            if args.shadow:
                shadow.apply_to_frames(cursor.frames, color=args.shadow_color, radius=args.shadow_radius,
                                       sigma=args.shadow_sigma, xoffset=args.shadow_x, yoffset=args.shadow_y)
            result = to_x11(cursor.frames)
            output = os.path.join(args.output, os.path.splitext(os.path.basename(name))[0])
            with open(output, 'wb') as f:
                f.write(result)

    with ThreadPool(cpu_count()) as pool:
        pool.map(process, args.files)


if __name__ == '__main__':
    main()



import os

from setuptools import setup, find_packages

with open(os.path.join(os.path.dirname(__file__), 'README.md')) as f:
    long_description = f.read()

setup(
    name='win2xcur',
    version='0.0.1',
    packages=find_packages(),
    install_requires=['Wand'],

    entry_points={
        'console_scripts': [
            'win2xcur = win2xcur.main:main',
        ],
    },

    author='quantum',
    author_email='quantum2048@gmail.com',
    url='https://github.com/quantum5/win2xcur',
    description='win2xcur is a tool to convert Windows .cur and .ani cursors to Xcursor format.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords='cur ani x11 windows win32 cursor xcursor',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Win32 (MS Windows)',
        'Environment :: X11 Applications',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Desktop Environment',
    ],
)



from typing import Iterator, List, Tuple

from wand.sequence import SingleImage


class CursorImage:
    def __init__(self, image: SingleImage, hotspot: Tuple[int, int]) -> None:
        self.image = image
        self.hotspot = hotspot

    def __repr__(self) -> str:
        return 'CursorImage(image=%r, hotspot=%r)' % (self.image, self.hotspot)


class CursorFrame:
    def __init__(self, images: List[CursorImage], delay=0) -> None:
        self.images = images
        self.delay = delay

    def __getitem__(self, item) -> CursorImage:
        return self.images[item]

    def __len__(self) -> int:
        return len(self.images)

    def __iter__(self) -> Iterator[CursorImage]:
        return iter(self.images)

    def __repr__(self) -> str:
        return 'CursorFrame(images=%r, delay=%r)' % (self.images, self.delay)
